﻿/**********************************************
 * Nobi's Somatosensory Game
 *  - Catch the Brick
 * 野比的体感游戏 - 抓住砖块 (B2B)
 * --------------------------------------------
 * 野比春熊
 * conmajia@gmail.com
 * May 13th, 2012 Mother's Day
 * --------------------------------------------
 * Visited my mom in the afternoon. Helped
 * setting up her Wii console which pushed
 * this somatosensory game idea into my mind.
 * 
 * Microsoft presented a powerful version:
 * the Kinect sensor(work with XBox360).
 * It has 2 visible light camera and 1 infrared
 * camera. That's VERY MUCH faster and smoother
 * than this little game.
 * If you feel this a piece of shit, try
 * Kinect, for only $200 (sensor only).
 * --------------------------------------------
 * Copyright 2012 Nobi Springbear as conmajia
 * You MAY ABLE TO USE, DUPLICATE this program.
 * But you MAY NOT MODIFY it WITHOUT my per-
 * mission.
 * (Ha ha, just a joke. You got this file, you
 * got everything already.)
 * Use as you wish. You can donate to me if
 * you feel grateful.
 * *******************************************/

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using Emgu.CV;
using Emgu.CV.Structure;
using System.Collections;
using SpringbearWorks.Image;

namespace NobiSomatosensoryGameBricks
{
    public partial class Form1 : Form
    {
        #region 字段和变量
        // debug
        bool debugMode = true;

        // game
        bool gameOn = false;

        // 缩放
        float xFactor = 1.0f;
        float yFactor = 1.0f;

        Random random = new Random();

        // 摄像头
        Capture camera;
        bool cameraInitiated = false;   // one-time use

        // 图像处理
        Image<Bgr, byte> sampleImage;
        Image<Bgr, byte> currentImage;
        Image<Bgr, byte> diffImage;
        Image<Bgr, byte> outputImage;
        int binaryThreshold = 0;

        #endregion

        #region 私有方法

        // 简单的碰撞检测
        Color pixelColor;
        BitmapEx bmp;

        private bool initiateCamera(int index)
        {
            if (camera != null)
                return false;

            try
            {
                camera = new Capture(index);
                return true;
            }
            catch (Exception ex)
            {
                Console.WriteLine("WEBCAM" + index + ":" + ex.Message);
                return false;
            }
        }

        private void showDialog(Control dialog)
        {
            if (dialog != null)
            {
                dialog.Visible = true;
                dialog.Dock = DockStyle.Fill;
            }
        }
        private void hideDialog(Control dialog)
        {
            if (dialog != null)
            {
                dialog.Visible = false;
                dialog.Dock = DockStyle.None;
            }
        }
        #endregion

        #region 初始化
        public Form1()
        {
            InitializeComponent();

            ball = new RectangleF(
                gamePictureBox.Width / 2,
                gamePictureBox.Height / 2,
                30,
                30);
        }

        private void Form1_Load(object sender, EventArgs e)
        {
        }
        void binaryThreshold_Click(object sender, EventArgs e)
        {
            binaryThreshold += 5;
            binaryThreshold %= 255;
            Console.WriteLine(binaryThreshold);
        }

        private void Form1_Activated(object sender, EventArgs e)
        {
            if (!cameraInitiated)
            {
                // init cam
                // no cam, no game
                bool cameraOk = initiateCamera(0);
                cameraButton.Enabled = cameraOk;
                cameraButton.Text =
                    "摄像头 ["
                    + (cameraOk ? "空闲" : "无效")
                    + "]";
                cameraInitiated = true;

                // 差异图
                if (diffImage != null)
                    diffImage.Dispose();
                diffImage = new Image<Bgr, byte>(camera.Width, camera.Height);
            }
        }
        #endregion

        #region 事件

        private void cameraButton_Click(object sender, EventArgs e)
        {
            // turn on/off camera and start the game
            if (!cameraInitiated)
            {
                cameraButton.Checked
                    = gameOn
                    = false;
                return;
            }

            if (cameraButton.Checked)
            {
                // on
                Application.Idle += new EventHandler(processGame);
            }
            else
            {
                // off
                Application.Idle -= new EventHandler(processGame);

                // clear
                gamePictureBox.Image = null;
            }

            gameOn = cameraButton.Checked;

            cameraButton.Text =
                "摄像头 ["
                + (gameOn ? "运行中" : "空闲")
                + "]";


            if (diffImage != null)
                bmp = new BitmapEx(
                    diffImage.Width,
                    diffImage.Height,
                    diffImage.Bitmap.PixelFormat);
            else
                bmp = new BitmapEx(
                    currentImage.Width,
                    currentImage.Height,
                    currentImage.Bitmap.PixelFormat);
        }

        private void exitButton_Click(object sender, EventArgs e)
        {
            // byebye
            if (gameOn)
            {
                cameraButton_Click(null, null);
            }
            Application.Exit();
        }

        private void aboutLabel_Click(object sender, EventArgs e)
        {
            aboutLabel.Visible = false;
            aboutLabel.Dock = DockStyle.None;
        }

        private void aboutButton_Click(object sender, EventArgs e)
        {
            if (aboutLabel.Visible)
                aboutLabel_Click(null, null);
            else
            {
                aboutLabel.Dock = DockStyle.Fill;
                aboutLabel.Visible = true;
            }
        }

        private void gameTimer_Tick(object sender, EventArgs e)
        {
        }

        #endregion

        #region 绘图
        // 测试用的蛋
        RectangleF ball;
        Color impact = Color.FromArgb(255, 0x08, 0x7b, 0x83);
        private void gamePictureBox_Paint(object sender, PaintEventArgs e)
        {
            if (gameOn)
            {

                //xFactor = (float)diffPictureBox.Width / (float)gamePictureBox.Width;
                //yFactor = (float)diffPictureBox.Height / (float)gamePictureBox.Height;


                // 检查碰撞 

                RectangleF tmp = new RectangleF();
                tmp.X = ball.X * xFactor;
                tmp.Y = ball.Y * yFactor;
                tmp.Width = ball.Width * xFactor;
                tmp.Height = ball.Height * yFactor;

                int whitePixelCount = 0;
                bmp._bitmap = diffImage.Bitmap;

                bmp.Lock();
                for (int i = 0; i < ball.Width; i++)
                {
                    for (int j = 0; j < ball.Height; j++)
                    {
                        pixelColor = bmp.GetPixel((int)tmp.X + i, (int)tmp.Y + j);
                        if (
                            (pixelColor.R + pixelColor.B + pixelColor.G)
                            >
                            2 * (impact.R + impact.G + impact.B)
                            )
                        {
                            whitePixelCount++;
                        }

                        currentImage.Bitmap.SetPixel((int)tmp.X + i, (int)tmp.Y + j, pixelColor);
                    }
                }
                bmp.Unlock();

                // more than 20% is conflict
                if (whitePixelCount > (0.2 * ball.Width * ball.Height))
                {
                    // conflict
                    // move egg to another place
                    ball.Location = new PointF(
                        random.Next(0, (int)(gamePictureBox.Width - ball.Width)),
                        random.Next(0, (int)(gamePictureBox.Height - ball.Height))
                        );
                }

                e.Graphics.DrawRectangle(Pens.Blue, tmp.X, tmp.Y, tmp.Width, tmp.Height);
                e.Graphics.FillRectangle(Brushes.Red, ball);
            }
        }
        #endregion

        #region 图像处理
        // 采集背景样本
        // 注意：背景上不要出现运动物体
        private void sampleButton_Click(object sender, EventArgs e)
        {
            sampleImage = currentImage;//.Convert<Gray, byte>();
        }

        // 比较图差异
        private Image<Bgr, byte> getDiff(Image<Bgr, byte> src, Image<Bgr, byte> dest)
        {
            // 输出二值化的差异图
            diffImage =
                currentImage
                .AbsDiff(sampleImage)
                .ThresholdBinary(
                new Bgr(Color.FromKnownColor((KnownColor)binaryThreshold)),
                new Bgr(255, 255, 255)
                );

            // 使用差异图作为蒙版去除原图背景
            return currentImage.Xor(diffImage);//.Xor(diffImage);
        }

        // process the game
        private void processGame(object sender, EventArgs e)
        {
            // process camera
            currentImage = camera.QueryFrame();
            // flip because my webcam is MIRRORED -- conmajia
            currentImage = currentImage.Flip(Emgu.CV.CvEnum.FLIP.HORIZONTAL);

            if (sampleImage != null)
            {
                // 计算图像差异
                diffImage = getDiff(currentImage, sampleImage);

            }

            gamePictureBox.Image = currentImage.ToBitmap(
                gamePictureBox.Width,
                gamePictureBox.Height);

        }
        #endregion
    }
}